#!/usr/bin/perl
use warnings FATAL => 'all';
use strict;
use Carp;
use Getopt::Long;
use File::Temp;
use File::Path;
use IO::Socket::INET;
use IO::Select;
use Config;
use Errno;

#
# Ensure that object destructors are called if this process is
# terminated by a signal.
#
use sigtrap qw(die normal-signals);

sub read_file($);
sub execute($);
sub execute_or_die($);
sub benchmark_commands($$$$);

#
# Attributes used by all sockets
#
my @common_socket_args = (
  Proto => 'tcp',
  Type => SOCK_STREAM,
  Timeout => 1,
  Blocking => 0,
);

#
# Command line parsing
#
my $storage_port;
my $storage_nodes;
my $config;
my $commands;
my $log = 'tahoe-bench.log';
my $tahoe = 'tahoe';
die "Could not parse command line options" unless GetOptions(
  "storage_port=i"  => \$storage_port,
  "storage_nodes=s" => \$storage_nodes,
  "config=s"        => \$config,	# Client template configuration file
  "commands=s"      => \$commands,	# File containing benchmark commands
  "log=s"           => \$log,		# Logfile to get benchmark results
  "tahoe=s"         => \$tahoe,		# Invoke the tahoe command line
);

my $i_am_storage = ($storage_port) ? 1 : 0;
my $i_am_client = ($storage_nodes and $config and $commands);
unless ($i_am_storage or $i_am_client) {
  print <<END;

 To run a storage node:
  -storage_port 	# Port to listen on for benchmark clients

 To run a client node and run benchmarks:
  -storage_nodes	# Text file, one address:port per line
  -config		# tahoe.cfg file to use

  -commands		# Text file of benchmark commands to run
     The following substitutions are made for each command:
      The string 'TAHOE' will be replaced with the tahoe invocation command
      The string 'NODE' will be replaced with the benchmark node directory
      The string 'ROOT' will be replaced with the benchmark root alias

  -log			# Name of logfile to append to
  -tahoe		# Command to invoke tahoe on this system

END
  exit 1;
}

print "Using this to invoke tahoe: $tahoe\n";

#
# If a storage_port has been defined, I will be a storage node.
# Create a socket to listen for connections on.
#
my $listener;
if ($i_am_storage) {
  $listener = new IO::Socket::INET(@common_socket_args,
      LocalPort => $storage_port, ReuseAddr => 1, Listen => 1);
  Carp::confess("Couldn't create socket: $!") unless $listener;
}

#
# If a storage_nodes file has been defined, read it.
# Try to connect to each storage node.
#
my @storage_sockets;
if ($i_am_client) {
  foreach my $line (read_file($storage_nodes)) {
    $line =~ s/#.*//;
    $line =~ s/^\s+//;
    $line =~ s/\s+$//;
    next if $line eq '';
    my ($addr, $port) = split /:/, $line;
    if (defined $addr and defined $port) {
      print "Trying $addr:$port\n";
      my $socket = new IO::Socket::INET(@common_socket_args,
	  PeerAddr => $addr, PeerPort => $port);
      if ($socket) {
	push @storage_sockets, $socket;
      } else {
	print "Couldn't connect to $addr:$port: $!\n";
	$! = 0;
      }
    } else {
      print "Ignoring line: $line\n";
    }
  }
}

while (1) {

  #
  # Storage node:  Wait for a connection
  #
  my $storage_socket;
  if ($i_am_storage) {
    while (not defined $storage_socket) {
      $storage_socket = $listener->accept();
      if (not defined $storage_socket and $! != Errno::ETIMEDOUT) {
	Carp::confess("accept(): $!");
      }
    }
    printf("Connected!\n");
  }

  #
  # Client node:  Start an introducer and broadcast its furl
  #
  my $introducer;
  if ($i_am_client) {
    $introducer = new introducer($tahoe);
    my $furl = $$introducer{furl};
    my $len = length($furl);
    foreach my $sock (@storage_sockets) {
      my $ret = send($sock, $furl, 0);
      if ($ret != $len) {
	Carp::confess("Error sending introducer to $sock");
      }
    }
  }

  #
  # Connected storage node:  Get the introducer.furl
  #
  my $storage_node;
  if ($i_am_storage) {
    my $line = '';
    my $got_newline = 0;
    while (not $got_newline) {
      sleep 1;
      my $temp;
      my $ret = $storage_socket->recv($temp, 255);
      if (defined $ret) {
	if (length($temp)) {
	  $line .= $temp;
	  $got_newline = $temp =~ /\n/;
	} else {
	  Carp::confess("Lost connection");
	}
      } else {
# 10035 A non-blocking socket operation could not be completed immediately
	Carp::confess("recv(): $!") if ($! != Errno::EAGAIN and $! != 10035);
      }
    }
    my ($furl) = $line =~ /^([^\n]+)\n/;
    print "Got introducer.furl = $furl\n";
    $storage_node = new storage($tahoe, $furl);
  }

  #
  # Client node:  Start a client and run the commands
  #
  if ($i_am_client) {
    my $client = new client($tahoe, $$introducer{furl}, $config);
    benchmark_commands($commands, $$client{dir}, $tahoe, $log);
    last;	# Benchmarking done
  }

  #
  # Connected storage node:  Wait for the connection to drop
  #
  while ($i_am_storage) {
    sleep 1;
    my $temp;
    my $ret = $storage_socket->recv($temp, 255);
    if (defined $ret) {
      if (length($temp)) {
	# I don't expect or care about any additional data
      } else {
	print "Disconnected!\n";
	last;	# Break out of the infinite loop
      }
    } else {
# 10035 A non-blocking socket operation could not be completed immediately
      Carp::confess("recv(): $!") if ($! != Errno::EAGAIN and $! != 10035);
    }
  }

}

# ----------------------------------------------------------------------------

#
# Return a list of all the lines from a text file
#
sub read_file($) {
  my ($file) = @_;
  my @lines;
  open(FILE, $file) or Carp::confess("Couldn't open '$file': $!");
  @lines = <FILE>;
  close(FILE) or Carp::confess("Couldn't close '$file': $!");
  return @lines;
}

# ----------------------------------------------------------------------------

#
# Execute an external program
#
# Return its output and an error string that is true if there was an error.
#
sub execute($) {
  my ($cmd) = @_;
  print "Executing $cmd\n";
  my $out = `$cmd 2>&1`;
  my $err = '';
  if (defined $out) {
    my $exit = 0xFF & ($? >> 8);
    my $sig  = 0x7F &  $?;
    my $core = 0x80 &  $?;
    $err .= "Command '$cmd' exited with status $exit\n" if $exit;
    $err .= "Command '$cmd' exited due to signal $sig\n" if $sig;
    $err .= "core dumped\n" if $core;
  } else {
    $out = '';	# Avoid undefined output string
    $err = "Couldn't execute '$cmd': $!\n";
  }
  return ($out, $err);
}

#
# Execute an external program
#
# Return its output if it succeeds, or end the script if it fails.
#
sub execute_or_die($) {
  my ($cmd) = @_;
  my ($out, $err) = execute($cmd);
  if ($err) {
    print STDERR $out;
    print STDERR $err;
    Carp::confess("Failure executing $cmd");
  }
  return $out;
}

#
# Benchmark the commands in a file, logging the results.
#
sub benchmark_commands($$$$) {
  my ($commands, $dir, $tahoe, $log) = @_;
  open(LOG, '>>', $log) or Carp::confess("Couldn't open '$log': $!");
  my $now = `date`;
  chomp $now;
  print LOG "$now - executing commands from $log\n";
  foreach my $command (read_file($commands)) {
    chomp $command;
    $command =~ s/TAHOE/$tahoe/;
    $command =~ s/NODE/$dir/;
    $command =~ s/ROOT/bench/;
    print LOG "Command: $command\n";
    my ($out, $err) = execute("/usr/bin/time -lp $command");
    #
    # The $err information isn't useful because it comes from /usr/bin/time
    # and not from the user's command.  So, throw it away.
    #
    print LOG $out;
  }
  close(LOG) or Carp::confess("Couldn't close '$log': $!");
}

# ----------------------------------------------------------------------------

package introducer;

sub new($$) {
  my ($class, $tahoe) = @_;
  my $self = {
    tahoe => $tahoe,
    dir   => File::Temp::tempdir("tahoe.XXXX", CLEANUP => 1, TMPDIR => 1),
    furl  => undef,
  };
  Carp::confess("Couldn't create tempdir: $!") unless defined $$self{dir};
  print main::execute_or_die("$tahoe create-introducer $$self{dir}");
  print main::execute_or_die("$tahoe start $$self{dir}");
  sleep 1;
  ($$self{furl}) = main::read_file("$$self{dir}/introducer.furl");
  return bless($self, $class);
}

sub DESTROY($) {
  my ($self) = @_;
  print main::execute_or_die("$$self{tahoe} stop $$self{dir}");
  File::Path::rmtree($$self{dir});
}

# ----------------------------------------------------------------------------

package storage;

#
# Create a tahoe storage node
#
# Use the specified $introducer and automatically create a nickname.
#
sub new($$$) {
  my ($class, $tahoe, $introducer) = @_;
  Carp::confess unless defined $introducer;
  my $self = {
    tahoe      => $tahoe,
    dir        => File::Temp::tempdir("tahoe.XXXX", CLEANUP => 1, TMPDIR => 1),
    introducer => $introducer,
  };
  Carp::confess("Couldn't create tempdir: $!") unless defined $$self{dir};
  print main::execute_or_die("$tahoe create-client $$self{dir}");
  my ($name) = $$self{dir} =~ /(tahoe.*)$/;
  my $file = "$$self{dir}/tahoe.cfg";
  open(FILE, ">$file") or Carp::confess("Couldn't open '$file': $!");
  print FILE <<END;
[node]
nickname = $name
web.port = tcp:4321
web.static = public_html
timeout.keepalive = 5
timeout.disconnect = 15
[client]
introducer.furl = $$self{introducer}
[storage]
enabled = true
reserved_space = 1MiB
expire.enabled = false
[helper]
enabled = false
END
  close(FILE) or Carp::confess("Couldn't close '$file': $!");
  print main::execute_or_die("$$self{tahoe} start $$self{dir}");
  sleep 1;
  return bless($self, $class);
}

sub DESTROY($) {
  my ($self) = @_;
  print main::execute_or_die("$$self{tahoe} stop $$self{dir}");
  File::Path::rmtree($$self{dir});
}

# ----------------------------------------------------------------------------

package client;

#
# Create a tahoe client, using the configuration file $config as a template.
#
# Use the specified $introducer and automatically create a nickname.
#
sub new($$$$) {
  my ($class, $tahoe, $introducer, $config) = @_;
  Carp::confess unless defined $introducer;
  Carp::confess unless defined $config;
  my $self = {
    tahoe      => $tahoe,
    dir        => File::Temp::tempdir("tahoe.XXXX", CLEANUP => 1, TMPDIR => 1),
    introducer => $introducer,
    config     => $config,
  };
  Carp::confess("Couldn't create tempdir: $!") unless defined $$self{dir};
  print main::execute_or_die("$tahoe create-client $$self{dir}");
  my ($name) = $$self{dir} =~ /(tahoe.*)$/;
  my $file = "$$self{dir}/tahoe.cfg";
  open(FILE, ">$file") or Carp::confess("Couldn't open '$file': $!");
  foreach my $line (main::read_file($$self{config})) {
    $line =~ s/nickname.*/nickname = $name/;
    $line =~ s/introducer\.furl.*/introducer.furl = $$self{introducer}/;
    print FILE $line;
  }
  close(FILE) or Carp::confess("Couldn't close '$file': $!");
  print main::execute_or_die("$$self{tahoe} start $$self{dir}");
  sleep 10;	# Give the client a few seconds to connect
  print main::execute_or_die("$$self{tahoe} create-alias -d $$self{dir} bench");
  return bless($self, $class);
}

sub DESTROY($) {
  my ($self) = @_;
  print main::execute_or_die("$$self{tahoe} stop $$self{dir}");
  File::Path::rmtree($$self{dir});
}